import type { BackendQueryFormat, BackendQueryResultFormat, GraphQueryResultMetaFromBackend, NodeStruct } from 'ts-common';
import { log } from '../../logger';


function applyFilter(op: string, value: number, threshold: number) {
  switch (op) {
    case '>': return value > threshold;
    case '<': return value < threshold;
    case '>=': return value >= threshold;
    case '<=': return value <= threshold;
    case '=': return value === threshold;
    case '!=': return value !== threshold;
    default: return true;
  }
}


export const runUdfs = async (
  actualResult: GraphQueryResultMetaFromBackend,
  query: BackendQueryFormat,
): Promise<GraphQueryResultMetaFromBackend> => {
  if (!query.udf) return actualResult;
  const result: GraphQueryResultMetaFromBackend = JSON.parse(JSON.stringify(actualResult));
  const loopThoughNodes = (node: NodeStruct): NodeStruct[] => {
    if (node.relation && node.relation.node) {
      return [node, ...loopThoughNodes(node.relation.node)];
    }
    return [node];
  };
  const allNodes = query.query.flatMap(path => loopThoughNodes(path.node));
  const allUniqueNodes: Record<
    string,
    { label: string; joins: { from: string; leftJoin: string; rightJoin: string; to: string; fromLabel: string; toLabel: string }[] }
  > = Object.fromEntries(
    allNodes
      .map(n => [n.id, { label: n.label, joins: [] }])
      .filter((node, index, self) => {
        return index === self.findIndex(t => t[0] === node[0] && t[1] === node[1]);
      }),
  );

  for (const udf of query.udf) {
    console.log('Running UDF', query.udf, allUniqueNodes);

    if (udf.type === 'degree') {
      const leftLabel = allUniqueNodes[udf.leftId].label;
      const rightLabel = udf.rightId ? allUniqueNodes[udf.rightId].label : undefined;

      const nodes = result.nodes;
      const edges = result.edges;

      const nodeMap: Record<string, number> = {};
      // Initialize the map with all nodes having a count of 0
      for (const node of nodes) {
        nodeMap[node._id] = 0;
      }
      // Count edges for each node
      for (let i = 0; i < edges.length; i++) {
        const edge = edges[i];
        nodeMap[edge.to] = nodeMap[edge.to] + 1;
        nodeMap[edge.from] = nodeMap[edge.from] + 1;
      }

      console.log('nodeMap', nodeMap, udf.right, query.udf);
      // Apply the operation filter based on degree
      //
      result.nodes = nodes.filter(node => {
        if (node.label !== leftLabel) return true;
        const degree = nodeMap[node._id] || 0;
        return applyFilter(udf.op, degree, udf.right)
      }
      );
      // Update edges to only include those connecting remaining nodes
      const nodeIds = new Set(result.nodes.map(node => node._id));
      result.edges = edges.filter(edge => nodeIds.has(edge.from) && nodeIds.has(edge.to));
    }

    else if (udf.type === 'community_detection') {
      console.log("performing community detection")
      class UnionFind {
        private parent: number[];
        private rank: number[];

        constructor(size: number) {
          this.parent = Array.from({ length: size }, (_, i) => i);
          this.rank = new Array(size).fill(0);
        }

        find(x: number): number {
          if (this.parent[x] !== x) {
            this.parent[x] = this.find(this.parent[x]); // Path compression
          }
          return this.parent[x];
        }

        union(x: number, y: number): void {
          const rootX = this.find(x);
          const rootY = this.find(y);
          if (rootX !== rootY) {
            // Union by rank
            if (this.rank[rootX] > this.rank[rootY]) {
              this.parent[rootY] = rootX;
            } else {
              this.parent[rootX] = rootY;
              if (this.rank[rootX] === this.rank[rootY]) this.rank[rootY]++;
            }
          }
        }
      }

      const nodes = result.nodes;
      const edges = result.edges;

      // Create node ID to index mapping
      const nodeIdToIndex: Record<string, number> = {};
      nodes.forEach((node, index) => {
        nodeIdToIndex[node._id] = index;
      });

      const uf = new UnionFind(nodes.length);

      // Process edges to build connected components
      edges.forEach(edge => {
        const fromIndex = nodeIdToIndex[edge.from];
        const toIndex = nodeIdToIndex[edge.to];
        uf.union(fromIndex, toIndex);
      });

      // adds cluster labels to each node
      const clusterMap: Record<number, number> = {};
      let currentClusterId = 0;
      nodes.forEach((node, index) => {
        const root = uf.find(index);
        if (clusterMap[root] === undefined) {
          clusterMap[root] = currentClusterId++;
        }
        node.cluster = clusterMap[root];

      });
    }
  }

  console.log(result.nodes.length, result.edges.length);
  return result;
};
